import os
import time
import unittest
import six
from mock import patch, MagicMock, Mock, mock_open
from urllib3.exceptions import SSLError

from txclib import utils, exceptions


# XXX: Taken from https://stackoverflow.com/a/21611963
def Any():
    """A method that will match any parameter."""
    class Any(object):
        def __eq__(self, *args):
            return True
    return Any()


class MakeRequestTestCase(unittest.TestCase):

    @patch('urllib3.PoolManager')
    def test_makes_request(self, mock_manager):
        response_mock = MagicMock()
        response_mock.status = 200
        response_mock.data = 'test_data'

        mock_connection = MagicMock()
        mock_connection.request.return_value = response_mock
        mock_manager.return_value = mock_connection

        host = 'http://whynotestsforthisstuff.com'
        url = '/my_test_url/'
        utils.make_request(
            'GET',
            host,
            url,
            'a_user',
            'a_pass'
        )
        mock_manager.assert_called_once_with(num_pools=1)
        mock_connection.request.assert_called_once()

        # In case the http proxy variable exists but is empty,
        # it shouldn't break things
        with patch.dict(os.environ, {'http_proxy': ''}):
            utils.make_request(
                'GET',
                host,
                url,
                'a_user',
                'a_pass'
            )
            mock_manager.assert_called_with(num_pools=1)
            self.assertEqual(mock_connection.request.call_count, 2)

        # In case the https proxy variable exists but is empty,
        # it shouldn't break things
        with patch.dict(os.environ, {'https_proxy': ''}):
            utils.make_request(
                'GET',
                host,
                url,
                'a_user',
                'a_pass'
            )
            mock_manager.assert_called_with(num_pools=1)
            self.assertEqual(mock_connection.request.call_count, 3)

    @patch('urllib3.ProxyManager')
    def test_makes_request_with_proxy(self, mock_manager):
        response_mock = MagicMock()
        response_mock.status = 200
        response_mock.data = 'test_data'

        mock_connection = MagicMock()
        mock_connection.request.return_value = response_mock
        mock_manager.return_value = mock_connection
        url = '/my_test_url/'

        # Test http
        host = 'http://whynotestsforthisstuff.com'
        with patch.dict(os.environ, {'http_proxy': 'proxy.host:333'}):
            utils.make_request(
                'GET',
                host,
                url,
                'a_user',
                'a_pass'
            )
            mock_manager.assert_called_once_with(num_pools=1,
                                                 proxy_headers=Any(),
                                                 proxy_url='http://proxy.host:333')
            mock_connection.request.assert_called_once()

        # Test https
        host = 'https://whynotestsforthisstuff.com'
        with patch.dict(os.environ, {'https_proxy': 'proxy.host:333'}):
            utils.make_request(
                'GET',
                host,
                url,
                'a_user',
                'a_pass'
            )
            mock_manager.assert_called_with(num_pools=1,
                                            proxy_headers=Any(),
                                            proxy_url='https://proxy.host:333',
                                            ca_certs=Any(),
                                            cert_reqs=Any())
            self.assertEqual(mock_connection.request.call_count, 2)

    @patch('urllib3.PoolManager')
    @patch('txclib.utils.logger')
    def test_catches_ssl_error(self, mock_logger, mock_manager):
        mock_connection = MagicMock()
        mock_connection.request.side_effect = SSLError('Boom!')
        mock_manager.return_value = mock_connection

        host = 'https://whynotestsforthisstuff.com'
        url = '/my_test_url/'
        self.assertRaises(
            SSLError,
            utils.make_request,
            'GET',
            host,
            url,
            'a_user',
            'a_pass'
        )
        mock_logger.error.assert_called_once_with("Invalid SSL certificate")

    @patch('txclib.utils.determine_charset')
    @patch('urllib3.PoolManager')
    def test_makes_request_skip_decode(self, mock_conn, mock_determine):
        response_mock = MagicMock()
        response_mock.status = 200
        response_mock.data = 'test_data'

        mock_connection = MagicMock()
        mock_connection.request.return_value = response_mock
        mock_conn.return_value = mock_connection

        host = 'http://whynotestsforthisstuff.com'
        url = '/my_test_url/'
        utils.make_request(
            'GET',
            host,
            url,
            'a_user',
            'a_pass',
            skip_decode=True
        )
        mock_conn.assert_called_once_with(num_pools=1)
        mock_connection.request.assert_called_once()
        mock_determine.assert_not_called()

    @patch('urllib3.PoolManager')
    def test_makes_request_404(self, mock_manager):
        response_mock = MagicMock()
        response_mock.status = 404
        response_mock.data = 'test_data'

        mock_connection = MagicMock()
        mock_connection.request.return_value = response_mock
        mock_manager.return_value = mock_connection

        host = 'http://whynotestsforthisstuff.com'
        url = '/my_test_url/'
        self.assertRaises(
            exceptions.HttpNotFound,
            utils.make_request,
            'GET',
            host,
            url,
            'a_user',
            'a_pass'
        )

    @patch('urllib3.PoolManager')
    def test_makes_request_403(self, mock_manager):
        response_mock = MagicMock()
        response_mock.status = 403
        response_mock.data = 'test_data'

        mock_connection = MagicMock()
        mock_connection.request.return_value = response_mock
        mock_manager.return_value = mock_connection

        host = 'http://whynotestsforthisstuff.com'
        url = '/my_test_url/'
        self.assertRaises(
            exceptions.HttpNotAuthorized,
            utils.make_request,
            'GET',
            host,
            url,
            'a_user',
            'a_pass'
        )

    @patch('urllib3.PoolManager')
    def test_makes_request_401(self, mock_manager):
        response_mock = MagicMock()
        response_mock.status = 401
        response_mock.data = 'test_data'

        mock_connection = MagicMock()
        mock_connection.request.return_value = response_mock
        mock_manager.return_value = mock_connection

        host = 'http://whynotestsforthisstuff.com'
        url = '/my_test_url/'
        self.assertRaises(
            exceptions.AuthenticationError,
            utils.make_request,
            'GET',
            host,
            url,
            'a_user',
            'a_pass'
        )

    @patch('urllib3.PoolManager')
    def test_makes_request_connection_error(self, mock_manager):
        """Tests for common 50X connection errors."""
        for code in range(500, 506):
            mock_connection = MagicMock()
            mock_connection.request.return_value = MagicMock(status=code,
                                                             data=None)
            mock_manager.return_value = mock_connection

            host = 'http://whynotestsforthisstuff.com'
            url = '/my_test_url/'
            args = ('GET', host, url, 'a_user', 'a_pass',)
            with self.assertRaises(exceptions.TXConnectionError) as err:
                utils.make_request(*args)
            self.assertEqual(err.exception.response_code, code)

    @patch('urllib3.PoolManager')
    def test_makes_request_None(self, mock_manager):
        response_mock = MagicMock()
        response_mock.status = 200
        response_mock.data = None

        mock_connection = MagicMock()
        mock_connection.request.return_value = response_mock
        mock_manager.return_value = mock_connection

        host = 'http://whynotestsforthisstuff.com'
        url = '/my_test_url/'
        utils.make_request(
            'GET',
            host,
            url,
            'a_user',
            'a_pass'
        )
        mock_manager.assert_called_once_with(num_pools=1)
        mock_connection.request.assert_called_once()

    @patch('urllib3.PoolManager')
    def test_url_format(self, mock_manager):
        response_mock = MagicMock()
        response_mock.status = 200
        response_mock.data = None

        mock_connection = MagicMock()
        mock_connection.request.return_value = response_mock
        mock_manager.return_value = mock_connection

        host = 'http://test.com/'
        url = '/path/to/we/'
        expected_url = 'http://test.com/path/to/we/'
        utils.make_request(
            'GET',
            host,
            url,
            'a_user',
            'a_pass'
        )
        mock_connection.request.assert_called_once_with('GET',
                                                        expected_url,
                                                        fields=Any(),
                                                        headers=Any())

    def test_get_current_branch_root_dir_no_git(self):
        with patch('txclib.utils.os.getcwd') as cwd_mock, \
             patch('txclib.utils.os.path.isdir') as isdir_mock:

            cwd_mock.return_value = '/usr/local/test_cli'
            isdir_mock.return_value = False
            b = utils.get_current_branch('/usr/local/test_cli')
            self.assertEqual(b, None)

    def test_get_current_branch_root_dir_with_git(self):
        data = "ref: refs/heads/test_branch\n"
        with patch('txclib.utils.os.getcwd') as cwd_mock, \
             patch('txclib.utils.os.path.isdir') as isdir_mock, \
             patch("txclib.utils.open", mock_open(read_data=data)):

            cwd_mock.return_value = '/usr/local/test_cli'
            isdir_mock.return_value = True
            b = utils.get_current_branch('/usr/local/test_cli')
            self.assertEqual(b, 'test_branch')

    def test_get_current_branch_subdir_with_git(self):
        data = "ref: refs/heads/test_branch\n"
        with patch('txclib.utils.os.getcwd') as cwd_mock, \
             patch('txclib.utils.os.path.isdir') as isdir_mock, \
             patch("txclib.utils.open", mock_open(read_data=data)):

            cwd_mock.return_value = '/usr/local/test_cli/files'
            isdir_mock.side_effects = [False, True]
            b = utils.get_current_branch('/usr/local/test_cli')
            self.assertEqual(b, 'test_branch')

    def test_get_current_branch_contains_slash(self):
        data = "ref: refs/heads/test_branch/abc\n"
        with patch('txclib.utils.os.getcwd') as cwd_mock, \
             patch('txclib.utils.os.path.isdir') as isdir_mock, \
             patch("txclib.utils.open", mock_open(read_data=data)):

            cwd_mock.return_value = '/usr/local/test_cli'
            isdir_mock.side_effects = [False, True]
            b = utils.get_current_branch('/usr/local/test_cli')
            self.assertEqual(b, 'test_branch/abc')


class ProjectFilesTestCase(unittest.TestCase):

    def test_project_files_scanning(self):
        """
        Test that the project files filter works as expected.
        """
        # XXX: This testcase depends on the current file structure

        # Test with a valid expression on directory-level
        expressions = ["tests/project_dir/test_expressions/<lang>/test.txt",
                       "./tests/project_dir/test_expressions/<lang>/test.txt"]

        for expr in expressions:
            langs = []
            for file, lang in utils.get_project_files(os.getcwd(), expr):
                langs.append(lang)
                self.assertTrue(file.endswith("{}/test.txt".format(lang)))
            self.assertListEqual(sorted(langs), sorted(["en", "es"]))

        # Test with a valid expression on file-level with different prefixes
        expressions = ["tests/project_dir/test_expressions/bulk/1.<lang>.po",
                       "tests/project_dir/test_expressions/bulk/2_<lang>.po",
                       "tests/project_dir/test_expressions/bulk/3 <lang>.po"]

        for expr in expressions:
            langs = []
            for file, lang in utils.get_project_files(os.getcwd(), expr):
                langs.append(lang)
                self.assertTrue(file.endswith("{}.po".format(lang)))
            self.assertListEqual(sorted(langs), sorted(["en_SE", "en"]))

        # Test with an invalid expression that doesn't contain <lang>
        expression = "tests/project_dir/test_expressions/bulk/1.en.po"
        msg = r"File filter (.*) does not contain"

        with six.assertRaisesRegex(self, exceptions.MalformedConfigFile, msg):
            for file, lang in utils.get_project_files(os.getcwd(), expression):
                pass

class GitUtilsTestCase(unittest.TestCase):

    def test_fetch_timestamp_from_git_tree(self):
        # A bit meta, lets try to get the timestamp of the
        # current file
        epoch_ts = utils.get_git_file_timestamp(
            os.path.dirname(os.path.abspath(__file__))
        )
        self.assertIsNotNone(epoch_ts)

    def test_git_timestamp_is_parsable(self):
        # A bit meta, lets try to get the timestamp of the
        # current file
        epoch_ts = utils.get_git_file_timestamp(
            os.path.dirname(os.path.abspath(__file__))
        )
        parsed_ts = time.mktime(time.gmtime(epoch_ts))
        self.assertIsNotNone(parsed_ts)

    def test_uses_authorized_date(self):
        commit = Mock()
        commit.authored_date = 1590969254
        commit.committed_date = 1590970456
        with patch('txclib.utils.git.Repo') as repo:
            repo.return_value.iter_commits.return_value = [commit]
            epoch_ts = utils.get_git_file_timestamp('any')
        self.assertEqual(1590969254, epoch_ts)



# -*- coding: utf-8 -*-
init_intro = """
 _____                    _  __
|_   _| __ __ _ _ __  ___(_)/ _| _____  __
  | || '__/ _` | '_ \/ __| | |_ / _ \ \/ /
  | || | | (_| | | | \__ \ |  _|  __/>  <
  |_||_|  \__,_|_| |_|___/_|_|  \___/_/\_\\


Welcome to the Transifex Client! Please follow the instructions to
initialize your project.
"""
init_initialized = "It seems that this project is already initialized."

init_reinit = "Do you want to delete it and reinit the project?"
init_host = "Transifex instance"

token_instructions = """
Transifex Client needs a Transifex API token to authenticate.
If you don’t have one yet, you can generate a token at
https://www.transifex.com/user/settings/api/.
"""

token_validation_failed = """
Error: Invalid token. You can generate a new token at
https://www.transifex.com/user/settings/api/.
"""
update_txrc = "Overwrite credentials in .transifexrc?"

token_msg = "[?] Enter your api token: "

running_tx_set = "Running tx config command for you..."


create_project_instructions = """To create a new project, head to {host}/{org}/add.
Once you’ve created the project, you can continue.
"""

TEXTS = {
    "source_file": {
        "description": ("""
The Transifex Client syncs files between your local directory and Transifex.
The mapping configuration between the two is stored in a file called .tx/config
in your current directory. For more information, visit
https://docs.transifex.com/client/config/.
"""),
        "error": ("No file was found in that path. Please correct the path "
                  "or make sure a file exists in that location."),
        "message": "[?] Enter the path to your local source file: "
    },
    "expression": {
        "description": ("""
Next, we’ll need a path expression pointing to the location of the
translation files (whether they exist yet or not) associated with
the source file ‘{source_file}’. You should include <lang> as a
wildcard for the language code.
"""),
        "error": "The path expression doesn’t contain the <lang> placeholder.",
        "message": "[?] Enter a path expression: "
    },
    "formats": {
        "description": ("""
Here’s a list of the supported file formats. For more information,
check our docs at https://docs.transifex.com/formats/.
"""),
        "error": "Error: Invalid choice. Enter the number corresponding with the format you wish to select.",
        "empty": "No formats found for this file extension. For more information, check our docs at https://docs.transifex.com/formats/",
        "message": "[?] Select the file format type [{r}]: ",
    },
    "organization": {
        "description": ("""
You’ll now choose a project in a Transifex organization to sync with your
local files. You belong to these organizations in Transifex:
"""),
        "error": ("""
Error: Invalid choice. Enter the number corresponding with the organization you wish to select."""),
        "message": "[?] Select the organization to use [{r}]: ",
    },
    "projects": {
        "description": ("""We found these projects in your organization."""),
        "error": """
Error: Invalid choice. Enter the number corresponding with the project you wish to select.""",
        "message": "[?] Select the project to use [{r}]: ",
    },
}

epilog = """
The Transifex Client syncs files between your local directory and Transifex.
The mapping configuration between the two is stored in a file called
.tx/config in your current directory. For more information, visit
https://docs.transifex.com/client/config/.
"""

final_instr = """
Here’s the content of the .tx/config file that was created:

    [{resource}]
    source_file = {source_file}
    file_filter = {file_filter}
    source_lang = {source_lang}
    type = {type}

You could also have generated the same configuration by running a single command like:

    tx config mapping -r {resource} -f {source_file} -s {source_lang} -t {type} '{file_filter}'

To learn more about the Config command, visit https://docs.transifex.com/client/config.

Here are some useful commands for your next steps:

Upload source files to Transifex:
    tx push  --source

Download translation files from Transifex once translations are done:
    tx pull --all
"""



from __future__ import unicode_literals
import functools
import os
import sys
import re
import git
import errno
import urllib3
import collections
import six
import platform
import txclib

try:
    from json import loads as parse_json
except ImportError:
    from simplejson import loads as parse_json
try:
    import configparser
except ImportError:
    import ConfigParser as configparser

try:
    from urlparse import urljoin  # Python 2
except ImportError:
    from urllib.parse import urljoin  # Python 3

from email.parser import Parser
from urllib3.exceptions import SSLError, HTTPError
from six.moves import input
from threading import Thread, current_thread, _MainThread
from txclib.urls import API_URLS
from txclib.exceptions import (
    UnknownCommandError, HttpNotFound, HttpNotAuthorized,
    AuthenticationError, TXConnectionError, MalformedConfigFile
)
from txclib.paths import posix_path, native_path, posix_sep
from txclib.web import user_agent_identifier, certs_file
from txclib.log import logger
from txclib.config import OrderedRawConfigParser, CERT_REQUIRED
from txclib.processors import visit_hostname


class ProjectNotInit(Exception):
    pass


DEFAULT_HOSTNAMES = {
    'hostname': 'https://www.transifex.com',
    'api_hostname': 'https://api.transifex.com'
}

THREADS = []


def get_base_dir():
    """PyInstaller Run-time Operation.

    http://pythonhosted.org/PyInstaller/#run-time-operation
    """
    if getattr(sys, 'frozen', False):
        # we are running in a bundle
        basedir = os.path.join(sys._MEIPASS, 'txclib')
    else:
        # we are running in a normal Python environment
        basedir = os.path.dirname(os.path.abspath(__file__))
    return basedir


def find_dot_tx(path=os.path.curdir, previous=None):
    """Return the path where .tx folder is found.

    The 'path' should be a DIRECTORY.
    This process is functioning recursively from the current directory to each
    one of the ancestors dirs.
    """
    path = os.path.abspath(path)
    if path == previous:
        return None
    joined = os.path.join(path, ".tx")
    if os.path.isdir(joined):
        return path
    else:
        return find_dot_tx(os.path.dirname(path), path)


#################################################
# Parse file filter expressions and create regex

def regex_from_filefilter(file_filter, root_path=os.path.curdir):
    """Create proper regex from <lang> expression."""
    # Force expr to be a valid regex expr (escaped) but keep <lang> intact
    expr_re = re.escape(
        posix_path(os.path.join(root_path, native_path(file_filter)))
    )
    expr_re = expr_re.replace("\\<lang\\>", '<lang>').replace(
        '<lang>', '([^%(sep)s]+)' % {'sep': re.escape(posix_sep)})

    return "^%s$" % expr_re


TX_URLS = collections.OrderedDict([
    ('resource', '(?P<hostname>https?://(\w|\.|:|-)+)/projects/p/(?P<project>(\w|-)+)/resource/(?P<resource>(\w|-)+)/?$'),  # noqa
    ('project', '(?P<hostname>https?://(\w|\.|:|-)+)/projects/p/(?P<project>(\w|-)+)/?$'),  # noqa
    ('project2', '(?P<hostname>https?://(\w|\.|:|-)+)/(\w|-)+/(?P<project>(\w|-)+)(/(dashboard|settings))?/?'),  # noqa
    ('resource2', '(?P<hostname>https?://(\w|\.|:|-)+)/(\w|-)+/(?P<project>(\w|-)+)/(?P<resource>(\w|-)+)/?'),  # noqa
])


def parse_tx_url(url):
    """
    Try to match given url to any of the valid url patterns specified in
    TX_URLS. If not match is found, we raise exception
    """
    for type_ in list(TX_URLS.keys()):
        pattern = TX_URLS[type_]
        m = re.match(pattern, url)
        if m:
            return type_, m.groupdict()
    raise Exception(
        "tx: Malformed URL given."
        " Please refer to our docs: http://bit.ly/txcconfig"
    )


def determine_charset(response):
    content_type = response.headers.get('content-type', None)
    if content_type:
        message = Parser().parsestr("Content-type: %s" % content_type)
        for charset in message.get_charsets():
            if charset:
                return charset
    return "utf-8"


def _prepare_url_request(host, username, password):
    """
    Return a ProxyManager object (as defined in urllib3 [1]) that can be used
    to perform authorized requests to a specific host.

    Authorization header is constructed and set using "username" and "password"
    parameters. Also set the common HTTP headers that we want to be sent with
    each request.

    [1]: http://urllib3.readthedocs.io/en/latest/reference/#urllib3.poolmanager.ProxyManager  # noqa
    """
    def get_proxy_url(env_proxy, scheme):
        parsed_proxy = urllib3.util.url.parse_url(env_proxy)
        return urllib3.util.url.Url(
            scheme=scheme,
            auth=parsed_proxy.auth,
            host=parsed_proxy.host,
            port=parsed_proxy.port,
            path=parsed_proxy.path,
            query=parsed_proxy.query,
            fragment=parsed_proxy.fragment
        )

    # Initialize http and https pool managers
    num_pools = 1
    managers = {}
    if host.lower().startswith("http://"):
        scheme = "http"
        if os.environ.get("http_proxy"):
            proxy_url = get_proxy_url(os.environ["http_proxy"], scheme)
            managers["http"] = urllib3.ProxyManager(
                proxy_url=proxy_url.url,
                proxy_headers=urllib3.util.make_headers(
                    user_agent=user_agent_identifier(),
                    proxy_basic_auth=proxy_url.auth),
                num_pools=num_pools
            )
        else:
            managers["http"] = urllib3.PoolManager(num_pools=num_pools)
    elif host.lower().startswith("https://"):
        scheme = "https"
        if os.environ.get("https_proxy"):
            proxy_url = get_proxy_url(os.environ["https_proxy"], scheme)
            managers["https"] = urllib3.ProxyManager(
                proxy_url=proxy_url.url,
                proxy_headers=urllib3.util.make_headers(
                    user_agent=user_agent_identifier(),
                    proxy_basic_auth=proxy_url.auth),
                num_pools=num_pools,
                cert_reqs=CERT_REQUIRED,
                ca_certs=certs_file()
            )
        else:
            managers["https"] = urllib3.PoolManager(
                num_pools=num_pools,
                cert_reqs=CERT_REQUIRED,
                ca_certs=certs_file()
            )
    else:
        raise Exception("Unknown scheme")

    headers = urllib3.util.make_headers(
        basic_auth='{0}:{1}'.format(username, password),
        accept_encoding=True,
        user_agent=user_agent_identifier(),
        keep_alive=True
    )

    manager = managers[scheme]

    return headers, manager


def queue_request(method, host, url, username, password, fields=None,
                  skip_decode=False, get_params=None, callback=None,
                  callback_args=None):
    """
    Add a request to the THREADS queue. Request will not be sent until the
    'perform_parallel_requests' method is called.
    """
    get_params = get_params or {}
    callback_args = callback_args or {}

    headers, manager = _prepare_url_request(host, username, password)
    # All arguments must be bytes, not unicode
    global THREADS
    THREADS.append(Thread(target=perform_single_request,
                          args=(method,
                                urljoin(host, url),
                                dict(headers),
                                fields,
                                manager,
                                skip_decode,
                                callback,
                                callback_args)))


def make_request(method, host, url, username, password, fields=None,
                 skip_decode=False, get_params=None, *args, **kwargs):
    """
    Perform a request.

    This is the (default) blocking method of making requests.
    """
    get_params = get_params or {}

    headers, manager = _prepare_url_request(host, username, password)
    # All arguments must be bytes, not unicode
    return perform_single_request(method,
                                  urljoin(host, url),
                                  dict(headers),
                                  fields,
                                  manager,
                                  skip_decode,
                                  *args,
                                  **kwargs)


def perform_parallel_requests():
    """
    Perform a set of requests in parallel using threads. The requests are saved
    in the global THREADS variable. We send requests in batches of 20 in order
    to achieve optimal performance without hitting API limits.
    """
    global THREADS

    if not THREADS:
        return

    total = len(THREADS)
    completed = 0
    update_progress(0, total)

    while THREADS:
        next_batch = THREADS[:10]
        THREADS = THREADS[10:]

        for thread in next_batch:
            thread.start()
        for thread in next_batch:
            thread.join()

        completed += len(next_batch)
        update_progress(completed, total)


def perform_single_request(method, url, headers, fields, manager, skip_decode,
                           callback=None, callback_args=None):
    callback_args = callback_args or {}
    response = None

    try:
        encoded_request = encode_args(manager.request)
        response = encoded_request(method, url, headers=headers, fields=fields)
        r_value = parse_tx_response(response, skip_decode)
    except SSLError:
        logger.error("Invalid SSL certificate")
        raise
    except HTTPError:
        logger.error("HTTP error")
        raise
    except Exception as e:
        logger.error(str(e))
        if isinstance(current_thread(), _MainThread):
            raise
        return
    finally:
        if response is not None:
            response.close()

    if callback is not None:
        callback_args.update({"data": r_value[0],
                              "charset": r_value[1]})
        callback(**callback_args)

    return r_value


def parse_tx_response(response, skip_decode):
    """
    Handle a response from the Transifex API.
    """
    charset = None
    data = response.data
    if not skip_decode:
        charset = determine_charset(response)
        if isinstance(data, bytes):
            data = data.decode(charset)
    if response.status < 200 or response.status >= 400:
        if response.status == 401:
            raise AuthenticationError(data)
        elif response.status == 403:
            raise HttpNotAuthorized(data)
        elif response.status == 404:
            raise HttpNotFound(data)
        elif response.status >= 500:
            msg = "Failed to connect. Server responded with HTTP code {}"
            raise TXConnectionError(msg.format(response.status),
                                    code=response.status)
        else:
            raise Exception("Error received from server: {}".format(data))

    return data, charset


def get_details(api_call, username, password, *args, **kwargs):
    """
    Get the tx project info through the API.
    This function can also be used to check the existence of a project.
    """
    url = API_URLS[api_call] % kwargs
    try:
        data, charset = make_request(
            'GET', kwargs['hostname'], url, username, password
        )
        return parse_json(data)
    except Exception as e:
        logger.debug(six.u(str(e)))
        raise


def valid_resource_slug(slug):
    """
    Check if a resource slug contains only valid characters.
    Valid format is [-_\w].[-_\w] (<project>.<resource>)
    """
    try:
        a, b = slug.split('.')
    except ValueError:
        return False
    else:
        return valid_slug(a) and valid_slug(b)


def valid_slug(slug):
    """
    Check if a slug contains only valid characters.
    Valid chars include [-_\w]
    """
    return re.match("^[A-Za-z0-9_-]+$", slug)


def discover_commands():
    """
    Inspect commands.py and find all available commands
    """
    import inspect
    from txclib import commands

    command_table = {}
    fns = inspect.getmembers(commands, inspect.isfunction)

    for name, fn in fns:
        if name.startswith("cmd_"):
            command_table.update({
                name.split("cmd_")[1]: fn
            })

    return command_table


def exec_command(command, *args, **kwargs):
    """
    Execute given command
    """
    commands = discover_commands()
    try:
        cmd_fn = commands[command]
    except KeyError:
        raise UnknownCommandError
    cmd_fn(*args, **kwargs)


def mkdir_p(path):
    try:
        if path:
            os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST:
            pass
        else:
            raise


def confirm(prompt='Continue?', default=True):
    """
    Prompt the user for a Yes/No answer.
    Args:
        prompt: The text displayed to the user ([Y/n] will be appended)
        default: If the default value will be yes or no
    """
    valid_yes = ['Y', 'y', 'Yes', 'yes', ]
    valid_no = ['N', 'n', 'No', 'no', ]
    if default:
        prompt = prompt + ' [Y/n]: '
        valid_yes.append('')
    else:
        prompt = prompt + ' [y/N]: '
        valid_no.append('')

    ans = input(prompt)
    while (ans not in valid_yes and ans not in valid_no):
        ans = input(prompt)

    return ans in valid_yes


# Stuff for command line colored output

COLORS = [
    'BLACK', 'RED', 'GREEN', 'YELLOW',
    'BLUE', 'MAGENTA', 'CYAN', 'WHITE'
]

DISABLE_COLORS = False


def color_text(text, color_name, bold=False):
    """
    This command can be used to colorize command line output. If the shell
    doesn't support this or the --disable-colors options has been set, it just
    returns the plain text.
    Usage:
        print "%s" % color_text("This text is red", "RED")
    """
    if color_name in COLORS and not DISABLE_COLORS:
        return '\033[%s;%sm%s\033[0m' % (
            int(bold), COLORS.index(color_name) + 30, text)
    else:
        return text


def _can_walk(base, subdirectories):
    """
    Determine if the first element of the 'subdirectories' list can be appended
    to the 'base' path and still specify a directory.
    """
    # Always consider the last part of the subdirectories the filename
    if len(subdirectories) < 2:
        return False
    # Allow any directory name that doesn't contain the <lang> placeholder
    return "<lang>" not in subdirectories[0]


def get_project_files(curpath, expression):
    """
    Iterate over the files in the project that match the given expression.
    Return a tuple with the absolute file path and the language code of the
    language that is associated with it.
    """
    # Strip the reference to the current directory, if it exists
    if expression.startswith(".{}".format(os.sep)):
        expression = expression[2:]
    # Split the expression into parts
    expression_parts = expression.split(os.sep)
    # Merge the expression's path into 'curpath' until the <lang> placeholder
    # is specified in order to reduce the search tree
    while _can_walk(curpath, expression_parts):
        curpath = os.path.realpath(os.path.join(curpath, expression_parts[0]))
        expression_parts = expression_parts[1:]
    expr_re = regex_from_filefilter(os.path.join(*expression_parts), curpath)
    expression_regex = re.compile(expr_re)

    initial_depth = curpath.count(os.sep)
    max_depth = 50
    for root, dirs, files in os.walk(curpath, followlinks=True):

        # Don't visit any subdirectory
        if root.count(os.sep) > initial_depth + max_depth:
            del dirs[:]
            continue

        for file_path in files:
            path_to_match = os.path.join(root, file_path)
            full_path = os.path.realpath(os.path.join(root, file_path))
            match = expression_regex.match(posix_path(path_to_match))
            if match:
                try:
                    lang = match.group(1)
                except IndexError:
                    msg = ("File filter '{}' does not contain the '<lang>' "
                           "placeholder".format(expression))
                    raise MalformedConfigFile(msg)
                yield full_path, lang



def encode_args(func):
    # we have to patch func in order to make tests work.
    # sadly mock does not have the attributes needed for functools
    # so we need to set the manually
    if not hasattr(func, '__name__'):
        functools.update_wrapper(func, str.split, ('__name__', ))

    @functools.wraps(func)
    def decorated(*args, **kwargs):
        new_args = _encode_anything(args)
        new_kwargs = _encode_anything(kwargs, keys_as_unicode=True)
        return func(*new_args, **new_kwargs)
    return decorated


def _encode_anything(thing, encoding='utf-8', keys_as_unicode=False):
    # Handle python versions
    if sys.version_info.major == 3:
        return thing

    if isinstance(thing, str):
        return thing
    elif isinstance(thing, unicode):
        return thing.encode(encoding)
    elif isinstance(thing, list):
        return [_encode_anything(item) for item in thing]
    elif isinstance(thing, tuple):
        return tuple(_encode_anything(list(thing)))
    elif isinstance(thing, dict):
        # I know this is weird, but when using kwargs in python-3, the keys
        # should be str, not bytes
        if keys_as_unicode:
            return {key: _encode_anything(value)
                    for key, value in thing.items()}
        else:
            return {_encode_anything(key): _encode_anything(value)
                    for key, value in thing.items()}
    elif thing is None:
        return thing
    else:
        raise TypeError("Could not encode, unknown type: {}".
                        format(type(thing)))


def get_config_file_path(root_path):
    """Check the .tx/config file exists."""
    config_file = os.path.join(root_path, ".tx", "config")
    logger.debug("Config file is %s" % config_file)
    if not os.path.exists(config_file):
        msg = "Cannot find the config file (.tx/config)!"
        raise ProjectNotInit(msg)
    return config_file


def get_tx_dir_path(path_to_tx):
    """Check the .tx directory exists."""
    root_path = path_to_tx or find_dot_tx()
    logger.debug("Path to tx is %s." % root_path)
    if not root_path:
        msg = "Cannot find any .tx directory!"
        raise ProjectNotInit(msg)
    return root_path


def read_config_file(config_file):
    """Parse the configuration file and return its contents."""
    config = OrderedRawConfigParser()
    try:
        config.read(config_file)
    except Exception as err:
        msg = "Cannot open/parse .tx/config file: %s" % err
        raise ProjectNotInit(msg)
    return config


def get_transifex_config(txrc_file):
    """Read the configuration from the .transifexrc files."""
    txrc = OrderedRawConfigParser()
    try:
        txrc.read((txrc_file,))
    except Exception as e:
        msg = "Cannot read configuration file: %s" % e
        raise ProjectNotInit(msg)
    migrate_txrc_file(txrc_file, txrc)
    return txrc


def migrate_txrc_file(txrc_file, txrc):
    """Migrate the txrc file, if needed."""
    if not os.path.exists(txrc_file):
        return txrc
    for section in txrc.sections():
        try:
            txrc.get(section, 'api_hostname')
        except configparser.NoOptionError:
            txrc.set(section, 'api_hostname',
                     DEFAULT_HOSTNAMES['api_hostname'])

        orig_hostname = txrc.get(section, 'hostname')
        hostname = visit_hostname(orig_hostname)
        if hostname != orig_hostname:
            msg = "Hostname %s should be changed to %s."
            logger.info(msg % (orig_hostname, hostname))
            if (sys.stdin.isatty() and sys.stdout.isatty() and
                    confirm('Change it now? ', default=True)):
                txrc.set(section, 'hostname', hostname)
                msg = 'Hostname changed'
                logger.info(msg)
            else:
                hostname = orig_hostname
        save_txrc_file(txrc_file, txrc)
    return txrc


def get_transifex_file(directory=None):
    """Fetch the path of the .transifexrc file.
    It is in the home directory of the user by default.
    """
    directory = find_dot_tx()
    if directory:
        local_txrc_file = os.path.join(directory, ".transifexrc")
        if os.path.exists(local_txrc_file):
            logger.debug(".transifexrc file is at %s" % directory)
            return local_txrc_file

    directory = os.path.expanduser('~')
    txrc_file = os.path.join(directory, ".transifexrc")
    logger.debug(".transifexrc file is at %s" % directory)
    if not os.path.exists(txrc_file):
        msg = "No credentials file was found at %s." % (txrc_file)
        logger.info(msg)
        mask = os.umask(0o077)
        open(txrc_file, 'w').close()
        os.umask(mask)
        if os.path.exists(txrc_file):
            logger.info('Created %s ' % txrc_file)
        else:
            logger.info('Could not create %s ' % txrc_file)
    return txrc_file


def save_tx_config(config_file, config):
    """Save the local configuration file."""
    fh = open(config_file, "w")
    config.write(fh)
    fh.close()


def save_txrc_file(txrc_file, txrc):
    """Save the .transifexrc file."""
    mask = os.umask(0o077)
    fh = open(txrc_file, 'w')
    txrc.write(fh)
    fh.close()
    os.umask(mask)


def get_api_domains(path_to_tx, host):
    try:
        txrc = get_transifex_config(get_transifex_file(path_to_tx))
        return {
            'hostname': txrc.get(host, 'hostname'),
            'api_hostname': txrc.get(host, 'api_hostname')
        }
    except (ProjectNotInit, configparser.NoOptionError,
            configparser.NoSectionError):
        return DEFAULT_HOSTNAMES


def get_current_branch(root_dir):
    """Searches for a .git directory starting from the
    current working directory and up until the project
    root.
    Return current branch if .git exists else None
    """
    found, git_dir = None, None
    cwd = os.getcwd()
    while root_dir in cwd:
        git_dir = os.path.join(cwd, '.git')
        if os.path.isdir(git_dir):
            found = True
            break
        cwd = os.path.dirname(cwd)

    if found:
        with open(os.path.join(git_dir, 'HEAD')) as f:
            ref = f.read()
            m = re.match('ref: refs/heads/(.+?)\s+$', ref)
            if m:
                return m.group(1)


def get_version():
    return '%s, py %s.%s, %s' % (
        txclib.__version__,
        sys.version_info.major,
        sys.version_info.minor,
        platform.machine()
    )


def update_progress(done, total):
    """
    Print a graphical progress bar to indicate a progress status.
    """
    percentage = 100.0 * done / float(total)
    template = '\r[{:25}] {:>3.0f}% ({}/{})'
    sys.stdout.write(template.format("#" * int(percentage/4),
                                     percentage, done, total))
    sys.stdout.flush()

    # Print a new line when done
    if done == total:
        sys.stdout.write('\n')


def get_git_file_timestamp(file_path):
    """
    Return the timestamp (epoch) for the latest commit of a file
    """
    try:
        repo = git.Repo()
        commits_touching_path = list(
            repo.iter_commits(paths=file_path, max_count=1)
        )
        if commits_touching_path:
            latest_commit = commits_touching_path[0]
            latest_commit_ts = latest_commit.authored_date
            return latest_commit_ts
        else:
            return None
    except git.InvalidGitRepositoryError:
        # Current path is not a git repo.
        return None



import os

from slugify import slugify
from txclib import messages
from txclib import utils
from txclib.api import Api
from txclib.project import Project
from txclib.log import logger
from six.moves import input


COLOR = "CYAN"

try:
    import readline
    readline.set_completer_delims(' \t\n;')
    readline.parse_and_bind("tab: complete")
except ImportError:
    pass


def validate_source_file(path):
    return os.path.isfile(os.path.abspath(path))


def validate_expression(expression):
    return '<lang>' in expression


def validate_int(choice, length):
    try:
        choice = int(choice)
    except ValueError:
        return False
    return 0 < choice <= length


def choice_prompt(l, key):
    """
    l: A list of tuples (key, display_value) with the valid choices
    key: one of 'formats', 'organizations', 'projects'
    returns the key of the selected choice
    """
    a = "\n".join(["  {}. {}".format(i+1, f[1])
                   for i, f in enumerate(l)])
    a = a + "\n"
    print(a)

    choice = ''
    first_time = True
    r = '1' if len(l) == 1 else '1-{}'.format(len(l))

    while not validate_int(choice, len(l)):
        if not first_time:
            print(messages.TEXTS[key]["error"])
        choice = input(utils.color_text(
            messages.TEXTS[key]['message'].format(r=r), COLOR))
        first_time = False
    return l[int(choice) - 1][0]


def input_prompt(key, validation_method):
    user_input = ''
    first_time = True
    while not validation_method(user_input):
        if not first_time:
            print(messages.TEXTS[key]['error'])

        user_input = input(
            utils.color_text(messages.TEXTS[key]['message'], COLOR))
        first_time = False
    return user_input


class Wizard(object):

    def __init__(self, path_to_tx):
        p = Project(path_to_tx)
        self.host = p.config.get('main', 'host')
        username, token_or_password = p.getset_host_credentials(
            self.host, only_token=True)

        self.api = Api(username=username, password=token_or_password,
                       host=self.host, path_to_tx=p.txrc_file)

    def get_organizations(self):
        try:
            organizations = self.api.get('organizations')
        except Exception as e:
            logger.error(e)
            raise

        # return org list sorted by name
        return sorted(
            [(o['slug'], o['name']) for o in organizations],
            key=lambda x: x[1]
        )

    def get_projects_for_org(self, organization):
        try:
            projects = self.api.get('projects', organization=organization)
        except Exception as e:
            logger.error(e)
            raise

        # return project list sorted by name
        return sorted(
            [p for p in projects if not p['archived']],
            key=lambda x: x['name']
        )

    def get_formats(self, filename):
        _, extension = os.path.splitext(filename)
        try:
            formats = self.api.get('formats')
        except Exception as e:
            logger.error(e)
            raise

        def display_format(v):
            return '{} - {}'.format(v['description'], v['file-extensions'])

        formats = [(k, display_format(v)) for k, v in formats.items()
                   if extension in v['file-extensions']]
        if not formats:
            raise Exception(messages.TEXTS['formats']['empty'])
        return sorted(formats, key=lambda x: x[0])

    def run(self):
        """
        Runs the interactive wizard for `tx set` command and populates the
        parser's options with the user input. Options `local` and `execute`
        are by default True when interactive wizard is run.

        Returns: the options dictionary.
        """

        TEXTS = messages.TEXTS

        print(TEXTS['source_file']['description'])
        source_file = input_prompt('source_file', validate_source_file)

        print(
            TEXTS['expression']['description'].format(source_file=source_file)
        )
        expression = input_prompt('expression', validate_expression)

        formats = self.get_formats(os.path.basename(source_file))
        print(TEXTS['formats']['description'])
        i18n_type = choice_prompt(formats, 'formats')

        organizations = self.get_organizations()
        print(TEXTS['organization']['description'])
        org_slug = choice_prompt(organizations, 'organization')

        projects = []
        first_time = True
        create_project = ("tx:new_project",
                          "Create new project (show instructions)...")
        project = None
        while not project:
            if not first_time:
                retry_message = "Hit Enter to try selecting a project again: "
                input(utils.color_text(retry_message, COLOR))

            projects = self.get_projects_for_org(org_slug)
            p_choices = [(p['slug'], p['name']) for p in projects]
            p_choices.append(create_project)
            if projects:
                print(TEXTS['projects']['description'])
            else:
                print("We found no projects in this organization!")
            first_time = False
            project_slug = choice_prompt(p_choices, 'projects')

            if project_slug == 'tx:new_project':
                print(messages.create_project_instructions.format(
                    host=self.host, org=org_slug
                ))
            else:
                project = [p for p in projects
                           if p['slug'] == project_slug][0]
                source_language = project['source_language']['code']

        resource_slug = slugify(os.path.basename(source_file))
        resource = '{}.{}'.format(project_slug, resource_slug)

        options = {
            'source_file': source_file,
            'expression': expression,
            'i18n_type': i18n_type,
            'source_language': source_language,
            'resource': resource,
        }
        return options



# -*- coding: utf-8 -*-

import os
import unittest
import itertools
try:
    import json
except ImportError:
    import simplejson as json
try:
    import configparser
except ImportError:
    import ConfigParser as configparser

from functools import wraps
from mock import Mock, patch, mock_open
from collections import namedtuple
from os.path import dirname
from sys import modules, version_info

from txclib.exceptions import AuthenticationError, TXConnectionError
from txclib.project import (Project, DEFAULT_PULL_URL)

from txclib.config import Flipdict
from txclib import utils


class TestProject(unittest.TestCase):

    @patch('txclib.utils.find_dot_tx')
    def test_get_tx_dir_path(self, m_find_dot_tx):
        """Test _get_tx_dir_path function"""
        expected_path = '/tmp/'
        m_find_dot_tx.return_value = expected_path
        path = utils.get_tx_dir_path(path_to_tx=None)
        self.assertEqual(path, expected_path)
        m_find_dot_tx.assert_called_once_with()

        expected_path = '/opt/'
        path = utils.get_tx_dir_path(path_to_tx=expected_path)
        self.assertEqual(path, expected_path)
        # make sure it has not been called twice
        m_find_dot_tx.assert_called_once_with()

    @patch('os.path.exists')
    def test_get_config_file_path(self, m_exists):
        """Test get_config_file_path function"""
        m_exists.return_value = True
        utils.get_config_file_path('/tmp/')
        m_exists.assert_called_once_with('/tmp/.tx/config')

        m_exists.return_value = False
        with self.assertRaises(utils.ProjectNotInit):
            utils.get_config_file_path('/tmp/')

    @patch('txclib.config.configparser')
    def test_getset_host_credentials(self, m_parser):
        p = Project(init=False)
        # let suppose a token has been set at the config
        dummy_token = 'salala'
        p.txrc = m_parser
        p.txrc.add_section = Mock()
        p.txrc.set = Mock()
        p.txrc.get = Mock()
        p.txrc.get.side_effect = ['api', dummy_token, None, None]
        p.txrc_file = '/tmp'
        username, password = p.getset_host_credentials('test')
        self.assertEqual(username, 'api')
        self.assertEqual(password, dummy_token)

        # let's try to get credentials for someone without
        # a token
        p.txrc.get.side_effect = [
            'username',
            'passw0rdz'
        ]
        username, password = p.getset_host_credentials('test')
        self.assertEqual(username, 'username')
        self.assertEqual(password, 'passw0rdz')

    @patch('txclib.project.input')
    @patch('txclib.config.configparser')
    def test_getset_host_credentials_no_transifexrc(
            self, m_parser, m_input):
        p = Project(init=False)
        # let suppose a token has been set at the config
        dummy_token = 'salala'
        p.txrc = m_parser
        p.save = Mock()
        p.validate_credentials = Mock(return_value=True)
        p.txrc_file = '/tmp'
        p.txrc.get.side_effect = configparser.NoSectionError('test')
        m_input.return_value = dummy_token
        username, password = p.getset_host_credentials('test')
        self.assertEqual(username, 'api')
        self.assertEqual(password, dummy_token)
        self.assertEqual(p.txrc.set.call_count, 2)
        self.assertEqual(m_input.call_count, 1)
        p.save.assert_called()

    @patch('txclib.project.input')
    @patch('txclib.config.configparser')
    @patch.dict(os.environ, {'TX_TOKEN': 'environment_value'})
    def test_getset_host_credentials_env_variable(
            self, m_parser, m_input):
        p = Project(init=False)
        p.txrc = m_parser
        p.save = Mock()
        p.txrc_file = '/tmp'
        p.txrc.has_section.side_effect = [False]
        username, password = p.getset_host_credentials('test')
        self.assertEqual(username, 'api')
        self.assertEqual(password, 'environment_value')
        self.assertEqual(p.txrc.set.call_count, 2)
        # no input will be asked, password will be used by environment variable
        self.assertEqual(m_input.call_count, 0)
        p.save.assert_called()

    @patch('txclib.project.input')
    @patch('txclib.config.configparser')
    @patch.dict(os.environ, {'TX_TOKEN': 'environment_value'})
    def test_getset_host_credentials_env_variable_first_time(
            self, m_parser, m_input):
        p = Project(init=False)
        p.txrc = m_parser
        p.save = Mock()
        p.txrc_file = '/tmp'
        p.txrc.has_section.side_effect = [False]
        username, password = p.getset_host_credentials('test')
        self.assertEqual(username, 'api')
        self.assertEqual(password, 'environment_value')
        # ensure that we have set host in the txrc_file, even though TX_TOKEN
        # exists
        self.assertEqual(p.txrc.set.call_count, 2)
        p.save.assert_called()

    @patch('txclib.config.configparser')
    @patch("txclib.project.logger")
    @patch.dict(os.environ, {'TX_TOKEN': 'environment_value'})
    def test_getset_host_credentials_both_token_and_env(
            self, m_logger, m_parser):
        p = Project(init=False)
        p.txrc = m_parser
        p.save = Mock()
        p.txrc_file = '/tmp'
        p.txrc.has_section.side_effect = [False]
        username, password = p.getset_host_credentials('test', token='demo')
        self.assertEqual(username, 'api')
        self.assertEqual(password, 'environment_value')
        # ensure that we did not make additional calls to set the token in the
        # txrc file
        self.assertEqual(p.txrc.set.call_count, 2)
        p.save.assert_called()
        self.assertEqual(m_logger.warning.call_count, 1)

    @patch('txclib.project.utils.confirm')
    @patch('txclib.config.configparser')
    def test_getset_host_credentials_update_transifexrc(
            self, m_parser, m_input):
        p = Project(init=False)
        dummy_token = 'salala'
        p.txrc = m_parser
        p.save = Mock()
        p.txrc_file = '/tmp'
        p.validate_credentials = Mock(return_value=True)
        p.txrc.get.side_effect = [
            'foo', 'bar'
        ]
        # transifexrc does not get updated if credentials are the same
        username, password = p.getset_host_credentials(
            'test', username='foo', password='bar'
        )
        self.assertEqual(username, 'foo')
        self.assertEqual(password, 'bar')
        self.assertEqual(p.txrc.set.call_count, 0)
        self.assertEqual(m_input.call_count, 0)
        self.assertEqual(p.save.call_count, 0)

        # transifexrc is not updated if confirm is no
        p.txrc.get.side_effect = [
            'foo', 'bar'
        ]
        m_input.return_value = False
        username, password = p.getset_host_credentials('test',
                                                       token=dummy_token)
        self.assertEqual(username, 'api')
        self.assertEqual(password, dummy_token)
        self.assertEqual(p.txrc.set.call_count, 0)
        self.assertEqual(m_input.call_count, 1)
        self.assertEqual(p.save.call_count, 0)

        # transifexrc is not updated if confirm is yes
        p.txrc.get.side_effect = [
            'foo', 'bar'
        ]
        m_input.return_value = True
        m_input.reset_mock()
        username, password = p.getset_host_credentials('test',
                                                       token=dummy_token)
        self.assertEqual(username, 'api')
        self.assertEqual(password, dummy_token)
        self.assertEqual(p.txrc.set.call_count, 2)
        self.assertEqual(m_input.call_count, 1)
        p.save.assert_called()

    def test_extract_fields(self):
        """Test the functions that extract a field from a stats object."""
        stats = {
            'completed': '80%',
            'last_update': '00:00',
            'foo': 'bar',
        }
        self.assertEqual(
            stats['completed'], '%s%%' % Project._extract_completed(stats)
        )
        self.assertEqual(stats['last_update'], Project._extract_updated(stats))

    def test_specifying_resources(self):
        """Test the various ways to specify resources in a project."""
        p = Project(init=False)
        resources = [
            'proj1.res1',
            'proj2.res2',
            'transifex.txn',
            'transifex.txo',
        ]
        with patch.object(p, 'get_resource_list') as mock:
            mock.return_value = resources
            cmd_args = [
                'proj1.res1', '*1*', 'transifex*', '*r*',
                '*o', 'transifex.tx?', 'transifex.txn',
            ]
            results = [
                ['proj1.res1', ],
                ['proj1.res1', ],
                ['transifex.txn', 'transifex.txo', ],
                ['proj1.res1', 'proj2.res2', 'transifex.txn', 'transifex.txo', ],  # noqa
                ['transifex.txo', ],
                ['transifex.txn', 'transifex.txo', ],
                ['transifex.txn', ],
                [],
            ]

            for i, arg in enumerate(cmd_args):
                resources = [arg]
                self.assertEqual(p.get_chosen_resources(resources), results[i])

            # wrong argument
            resources = ['*trasnifex*', ]
            self.assertRaises(Exception, p.get_chosen_resources, resources)


class TestProjectMinimumPercent(unittest.TestCase):
    """Test the minimum-perc option."""

    def setUp(self):
        super(TestProjectMinimumPercent, self).setUp()
        self.p = Project(init=False)
        self.p.minimum_perc = None
        self.p.resource = "resource"

    def test_cmd_option(self):
        """Test command-line option."""
        self.p.minimum_perc = 20
        results = itertools.cycle([80, 90])

        def side_effect(*args):
            return next(results)

        with patch.object(self.p, "get_resource_option") as mock:
            mock.side_effect = side_effect
            self.assertFalse(
                self.p._satisfies_min_translated({'completed': '12%'})
            )
            self.assertTrue(
                self.p._satisfies_min_translated({'completed': '20%'})
            )
            self.assertTrue(
                self.p._satisfies_min_translated({'completed': '30%'})
            )

    def test_global_only(self):
        """Test only global option."""
        results = itertools.cycle([80, None])

        def side_effect(*args):
            return next(results)

        with patch.object(self.p, "get_resource_option") as mock:
            mock.side_effect = side_effect
            self.assertFalse(
                self.p._satisfies_min_translated({'completed': '70%'})
            )
            self.assertTrue(
                self.p._satisfies_min_translated({'completed': '80%'})
            )
            self.assertTrue(
                self.p._satisfies_min_translated({'completed': '90%'})
            )

    def test_local_lower_than_global(self):
        """Test the case where the local option is lower than the global."""
        results = itertools.cycle([80, 70])

        def side_effect(*args):
            return next(results)

        with patch.object(self.p, "get_resource_option") as mock:
            mock.side_effect = side_effect
            self.assertFalse(
                self.p._satisfies_min_translated({'completed': '60%'})
            )
            self.assertTrue(
                self.p._satisfies_min_translated({'completed': '70%'})
            )
            self.assertTrue(
                self.p._satisfies_min_translated({'completed': '80%'})
            )
            self.assertTrue(
                self.p._satisfies_min_translated({'completed': '90%'})
            )

    def test_local_higher_than_global(self):
        """Test the case where the local option is lower than the global."""
        results = itertools.cycle([60, 70])

        def side_effect(*args):
            return next(results)

        with patch.object(self.p, "get_resource_option") as mock:
            mock.side_effect = side_effect
            self.assertFalse(
                self.p._satisfies_min_translated({'completed': '60%'})
            )
            self.assertTrue(
                self.p._satisfies_min_translated({'completed': '70%'})
            )
            self.assertTrue(
                self.p._satisfies_min_translated({'completed': '80%'})
            )
            self.assertTrue(
                self.p._satisfies_min_translated({'completed': '90%'})
            )

    def test_local_only(self):
        """Test the case where the local option is lower than the global."""
        results = itertools.cycle([None, 70])

        def side_effect(*args):
            return next(results)

        with patch.object(self.p, "get_resource_option") as mock:
            mock.side_effect = side_effect
            self.assertFalse(
                self.p._satisfies_min_translated({'completed': '60%'})
            )
            self.assertTrue(
                self.p._satisfies_min_translated({'completed': '70%'})
            )
            self.assertTrue(
                self.p._satisfies_min_translated({'completed': '80%'})
            )
            self.assertTrue(
                self.p._satisfies_min_translated({'completed': '90%'})
            )

    def test_no_option(self):
        """"Test the case there is nothing defined."""
        results = itertools.cycle([None, None])

        def side_effect(*args):
            return next(results)

        with patch.object(self.p, "get_resource_option") as mock:
            mock.side_effect = side_effect
            self.assertTrue(
                self.p._satisfies_min_translated({'completed': '0%'})
            )
            self.assertTrue(
                self.p._satisfies_min_translated({'completed': '10%'})
            )
            self.assertTrue(
                self.p._satisfies_min_translated({'completed': '90%'})
            )


class TestProjectFilters(unittest.TestCase):
    """
    Test filters used to decide whether to push/pull a translation or not.
    """

    def setUp(self):
        super(TestProjectFilters, self).setUp()
        self.p = Project(init=False)
        self.p.minimum_perc = None
        self.p.resource = "resource"
        self.stats = {
            'en': {
                'completed': '100%', 'last_update': '2011-11-01 15:00:00',
            }, 'el': {
                'completed': '60%', 'last_update': '2011-11-01 15:00:00',
            }, 'pt': {
                'completed': '70%', 'last_update': '2011-11-01 15:00:00',
            },
        }
        self.langs = list(self.stats.keys())

    def test_add_translation(self):
        """Test filters for adding translations.

        We do not test here for minimum percentages.
        """
        with patch.object(self.p, "get_resource_option") as mock:
            mock.return_value = None
            should_add = self.p._should_add_translation
            for force in [True, False]:
                for lang in self.langs:
                    self.assertTrue(should_add(lang, self.stats, force))

            # unknown language
            self.assertFalse(should_add('es', self.stats))

    def test_update_translation(self):
        """Test filters for updating a translation.

        We do not test here for minimum percentages.
        """
        with patch.object(self.p, "get_resource_option") as mock:
            mock.return_value = None

            should_update = self.p._should_update_translation
            force = True
            for lang in self.langs:
                self.assertTrue(should_update(lang, self.stats, 'foo', force))

            force = False       # reminder
            local_file = 'foo'

            # unknown language
            self.assertFalse(should_update('es', self.stats, local_file))

            # no local file
            with patch.object(self.p, "_get_time_of_local_file") as time_mock:
                time_mock.return_value = None
                with patch.object(self.p, "get_full_path") as path_mock:
                    path_mock.return_value = "foo"
                    for lang in self.langs:
                        self.assertTrue(
                            should_update(lang, self.stats, local_file)
                        )

            # older local files
            local_times = [self.p._generate_timestamp('2011-11-01 14:00:59')]
            results = itertools.cycle(local_times)

            def side_effect(*args):
                return next(results)

            with patch.object(self.p, "_get_time_of_local_file") as time_mock:
                time_mock.side_effect = side_effect
                with patch.object(self.p, "get_full_path") as path_mock:
                    path_mock.return_value = "foo"
                    for lang in self.langs:
                        self.assertTrue(
                            should_update(lang, self.stats, local_file)
                        )

            # newer local files
            local_times = [self.p._generate_timestamp('2011-11-01 15:01:59')]
            results = itertools.cycle(local_times)

            def side_effect(*args):
                return next(results)

            with patch.object(self.p, "_get_time_of_local_file") as time_mock:
                time_mock.side_effect = side_effect
                with patch.object(self.p, "get_full_path") as path_mock:
                    path_mock.return_value = "foo"
                    for lang in self.langs:
                        self.assertFalse(
                            should_update(lang, self.stats, local_file)
                        )

    def test_push_translation(self):
        """
        Test filters for pushing a translation file.
        """
        with patch.object(self.p, "get_resource_option") as mock:
            mock.return_value = None

            local_file = 'foo'
            should_push = self.p._should_push_translation
            force = True
            for lang in self.langs:
                self.assertTrue(
                    should_push(
                        lang, self.stats, local_file, force
                    )
                )

            force = False  # reminder

            # unknown language
            self.assertTrue(should_push('es', self.stats, local_file))

            # older local files
            local_times = [self.p._generate_timestamp('2011-11-01 14:00:59')]
            results = itertools.cycle(local_times)

            def side_effect(*args):
                return next(results)

            with patch.object(self.p, "_get_time_of_local_file") as time_mock:
                time_mock.side_effect = side_effect
                with patch.object(self.p, "get_full_path") as path_mock:
                    path_mock.return_value = "foo"
                    for lang in self.langs:
                        self.assertFalse(
                            should_push(lang, self.stats, local_file)
                        )

            # newer local files
            local_times = [self.p._generate_timestamp('2011-11-01 15:01:59')]
            results = itertools.cycle(local_times)

            def side_effect(*args):
                return next(results)

            with patch.object(self.p, "_get_time_of_local_file") as time_mock:
                time_mock.side_effect = side_effect
                with patch.object(self.p, "get_full_path") as path_mock:
                    path_mock.return_value = "foo"
                    for lang in self.langs:
                        self.assertTrue(
                            should_push(lang, self.stats, local_file)
                        )


class TestProjectPull(unittest.TestCase):
    """Test bits & pieces of the pull method."""

    def setUp(self):
        super(TestProjectPull, self).setUp()
        self.p = Project(init=False)
        self.p.minimum_perc = None
        self.p.resource = "resource"
        self.p.host = 'foo'
        self.p.project_slug = 'foo'
        self.p.resource_slug = 'foo'
        self.stats = {
            'en': {
                'completed': '100%', 'last_update': '2011-11-01 15:00:00',
            }, 'el': {
                'completed': '60%', 'last_update': '2011-11-01 15:00:00',
            }, 'pt': {
                'completed': '70%', 'last_update': '2011-11-01 15:00:00',
            },
        }
        self.langs = list(self.stats.keys())
        self.files = dict(list(zip(self.langs, itertools.repeat(None))))
        self.details = {'available_languages': []}
        for lang in self.langs:
            self.details['available_languages'].append({'code': lang})
        self.slang = 'en'
        self.lang_map = Flipdict()

    def test_new_translations(self):
        """Test finding new translations to add."""
        with patch.object(self.p, 'do_url_request') as resource_mock:
            resource_mock.return_value = json.dumps(self.details), "utf-8"
            new_trans = self.p._new_translations_to_add
            for force in [True, False]:
                res = new_trans(
                    self.files, self.slang, self.lang_map, self.stats, force
                )
                self.assertEqual(res, set([]))

            with patch.object(self.p,
                              '_should_add_translation') as filter_mock:
                filter_mock.return_value = True
                for force in [True, False]:
                    res = new_trans(
                        {'el': None}, self.slang, self.lang_map, self.stats,
                        force
                    )
                    self.assertEqual(res, set(['pt']))
                for force in [True, False]:
                    res = new_trans(
                        {}, self.slang, self.lang_map, self.stats, force
                    )
                    self.assertEqual(res, set(['el', 'pt']))

                files = {}
                files['pt_PT'] = None
                lang_map = {'pt': 'pt_PT'}
                for force in [True, False]:
                    res = new_trans(
                        files, self.slang, lang_map, self.stats, force
                    )
                    self.assertEqual(res, set(['el']))

    def test_get_pseudo_file(self):
        slang = 'en'
        resource = 'adriana'
        file_filter = 'adriana/<lang>.po'

        pseudo_file = self.p._get_pseudo_file(slang, resource, file_filter)

        self.assertEqual(pseudo_file, 'adriana/en_pseudo.po')

    def test_languages_to_pull_empty_initial_list(self):
        """Test determining the languages to pull, when the initial
        list is empty.
        """
        languages = []
        force = False

        res = self.p._languages_to_pull(
            languages, self.files, self.lang_map, self.stats, force
        )
        existing = res[0]
        new = res[1]
        self.assertEqual(existing, set(['el', 'en', 'pt']))
        self.assertFalse(new)

        del self.files['el']
        self.files['el-gr'] = None
        self.lang_map['el'] = 'el-gr'
        res = self.p._languages_to_pull(
            languages, self.files, self.lang_map, self.stats, force
        )
        existing = res[0]
        new = res[1]
        self.assertEqual(existing, set(['el', 'en', 'pt']))
        self.assertFalse(new)

    def test_languages_to_pull_with_initial_list(self):
        """Test determining the languages to pull, then there is a
        language selection from the user.
        """
        languages = ['el', 'en']
        self.lang_map['el'] = 'el-gr'
        del self.files['el']
        self.files['el-gr'] = None
        force = False

        with patch.object(self.p, '_should_add_translation') as mock:
            mock.return_value = True
            res = self.p._languages_to_pull(
                languages, self.files, self.lang_map, self.stats, force
            )
            existing = res[0]
            new = res[1]
            self.assertEqual(existing, set(['en', 'el-gr', ]))
            self.assertFalse(new)

            mock.return_value = False
            res = self.p._languages_to_pull(
                languages, self.files, self.lang_map, self.stats, force
            )
            existing = res[0]
            new = res[1]
            self.assertEqual(existing, set(['en', 'el-gr', ]))
            self.assertFalse(new)

            del self.files['el-gr']
            mock.return_value = True
            res = self.p._languages_to_pull(
                languages, self.files, self.lang_map, self.stats, force
            )
            existing = res[0]
            new = res[1]
            self.assertEqual(existing, set(['en', ]))
            self.assertEqual(new, set(['el', ]))

            mock.return_value = False
            res = self.p._languages_to_pull(
                languages, self.files, self.lang_map, self.stats, force
            )
            existing = res[0]
            new = res[1]
            self.assertEqual(existing, set(['en', ]))
            self.assertEqual(new, set([]))

    def test_in_combination_with_force_option(self):
        """Test the minimum-perc option along with -f."""
        with patch.object(self.p, 'get_resource_option') as mock:
            mock.return_value = 70

            res = self.p._should_download('de', self.stats, None, False)
            self.assertEqual(res, False)
            res = self.p._should_download('el', self.stats, None, False)
            self.assertEqual(res, False)
            res = self.p._should_download('el', self.stats, None, True)
            self.assertEqual(res, False)
            res = self.p._should_download('en', self.stats, None, False)
            self.assertEqual(res, True)
            res = self.p._should_download('en', self.stats, None, True)
            self.assertEqual(res, True)

            with patch.object(self.p, '_remote_is_newer'):
                res = self.p._should_download('pt', self.stats, None, False)
                self.assertEqual(res, True)
                res = self.p._should_download('pt', self.stats, None, True)
                self.assertEqual(res, True)

            with patch.object(utils, 'get_git_file_timestamp') as ts_mock:
                # Note that the test needs an existing file path
                # The current test (__file__) is the only file
                # we're sure will exist and won't change

                # Old timestamp (1990)
                ts_mock.return_value = 640124371
                res = self.p._should_download(
                    'pt', self.stats, os.path.abspath(__file__), False,
                    use_git_timestamps=True
                )
                self.assertEqual(res, True)

                # "Recent" timestamp (in the future - 2100)
                ts_mock.return_value = 4111417171
                res = self.p._should_download(
                    'pt', self.stats, os.path.abspath(__file__), False,
                    use_git_timestamps=True
                )
                self.assertEqual(res, False)

    def test_get_url_by_pull_mode(self):
        self.assertEqual(
            'pull_sourceastranslation_file',
            self.p._get_url_by_pull_mode(mode='sourceastranslation')
        )
        self.assertEqual(
            DEFAULT_PULL_URL,
            self.p._get_url_by_pull_mode(mode='invalid mode')
        )
        self.assertEqual(
            DEFAULT_PULL_URL,
            self.p._get_url_by_pull_mode(mode=None)
        )

    def fixture_mocked_project(func):
        """A mock object with main os and http operations mocked"""
        @wraps(func)
        def wrapper(*args, **kwargs):
            app_dir = dirname(modules['txclib'].__file__)
            config_file = app_dir + "/../tests/templates/config"
            transifex_file = app_dir + "/../tests/templates/transifexrc"
            with patch("txclib.utils.encode_args") as mock_encode_args, \
                patch("txclib.utils.determine_charset")\
                    as mock_determine_charset, \
                    patch("txclib.utils.get_transifex_file",
                          return_value=transifex_file) \
                    as mock_get_transifex_file, \
                    patch("txclib.utils.get_config_file_path",
                          return_value=config_file) \
                    as mock_get_config_file_path, \
                    patch("txclib.utils.save_txrc_file") \
                    as mock_save_txrc_file, \
                    patch("txclib.project.Project._get_stats_for_resource") \
                    as mock_get_stats_for_resource:

                # Create fake https response
                def encode_args(*args, **kwargs):
                    struct = namedtuple("response", "data status close")
                    return struct(status=401, data="mock_response",
                                  close=Mock())
                mock_determine_charset.return_value = "utf-8"
                mock_encode_args.return_value = encode_args

                # Mock configuration files
                p = Project(init=False)
                p._init(path_to_tx=app_dir + "/../templates")

                kwargs['mock_project'] = p
                kwargs['mocks'] = {
                    'mock_determine_charset': mock_determine_charset,
                    "mock_encode_args": mock_encode_args,
                    "mock_get_config_file_path": mock_get_config_file_path,
                    "mock_get_stats_for_resource": mock_get_stats_for_resource,
                    "mock_get_transifex_file": mock_get_transifex_file,
                    "mock_save_txrc_file": mock_save_txrc_file
                }
                return func(*args, **kwargs)
        return wrapper

    @fixture_mocked_project
    @patch("txclib.project.logger.error")
    def test_pull_raises_authentication_exception(self, mock_logger, **kwargs):
        project = kwargs['mock_project']
        with self.assertRaises(AuthenticationError):
            project.pull()
            mock_logger.assert_called_once_with(
                Project.AUTHENTICATION_FAILED_MESSAGE)

    @fixture_mocked_project
    @patch('txclib.project.Project.do_url_request')
    def test_pull_with_branch_pulls_from_right_resource(self, m, **kwargs):
        m.return_value = ('{"i18n_type": "PO"}', '')
        project = kwargs['mock_project']
        project.pull(branch='somebranch')
        self.assertDictEqual(project.url_info, {
            'host': 'https://fake.com',
            'project': 'example',
            'resource': 'somebranch--enpo'
        })

    @fixture_mocked_project
    @patch('txclib.project.Project.do_url_request')
    def test_push_with_branch_pushes_to_right_resource(self, m, **kwargs):
        m.return_value = ('{"i18n_type": "PO"}', '')
        project = kwargs['mock_project']
        project.push(source=True, branch='somebranch')
        self.assertDictEqual(project.url_info, {
            'host': 'https://fake.com',
            'project': 'example',
            'resource': 'somebranch--enpo'
        })

    @fixture_mocked_project
    @patch('txclib.project.Project.do_url_request')
    def test_push_with_branch_weird_characters_are_handled(self, m, **kwargs):
        m.return_value = ('{"i18n_type": "PO"}', '')
        project = kwargs['mock_project']
        project.push(source=True, branch='some/b**r:a&n!ch')
        self.assertDictEqual(project.url_info, {
            'host': 'https://fake.com',
            'project': 'example',
            'resource': 'some-b-r-a-n-ch--enpo'
        })

    @fixture_mocked_project
    @patch("txclib.project.logger.error")
    def test_push_raises_authentication_exception(self, mock_logger, **kwargs):
        project = kwargs['mock_project']
        with self.assertRaises(AuthenticationError):
            project.push()

    @fixture_mocked_project
    @patch("txclib.project.logger.error")
    def test_delete_raises_authentication_exception(self, mock_logger,
                                                    **kwargs):
        project = kwargs['mock_project']
        with self.assertRaises(AuthenticationError):
            project.delete()

    @fixture_mocked_project
    @patch("txclib.project.logger.error")
    @patch("txclib.utils.make_request")
    def test_pull_raises_connection_exception(self, mock_request, mock_logger,
                                              **kwargs):
        """Test that all connection errors are properly handled."""
        project = kwargs["mock_project"]
        response = 502
        msg = "Failed with code %d" % response
        mock_request.side_effect = TXConnectionError(msg, code=response)
        with self.assertRaises(TXConnectionError):
            project.pull()

    @fixture_mocked_project
    @patch("txclib.project.logger.error")
    @patch("txclib.utils.make_request")
    def test_push_raises_connection_exception(self, mock_request, mock_logger,
                                              **kwargs):
        """Test that all connection errors are properly handled."""
        project = kwargs["mock_project"]
        response = 500
        msg = "Failed with code %d" % response
        mock_request.side_effect = TXConnectionError(msg, code=response)
        with self.assertRaises(TXConnectionError):
            project.push()

    @fixture_mocked_project
    @patch('txclib.utils.queue_request')
    @patch('txclib.utils.make_request')
    @patch('txclib.project.Project._resource_exists')
    def test_push_async(self, mock_exists, mock_request, mock_queue, **kwargs):
        mock_queue.return_value = None
        mock_request.return_value = ('{"i18n_type": "PO"}', '')
        mock_exists.return_value = True
        project = kwargs['mock_project']
        patch_open = ("builtins.open" if version_info.major > 2
                      else "__builtin__.open")
        with patch(patch_open, mock_open(read_data="")):
            project.push(source=True, parallel=True)
            self.assertEqual(mock_queue.call_count, 1)

            project.push(translations=True, parallel=True)
            self.assertEqual(mock_queue.call_count, 1)

            project.pull(parallel=True)
            self.assertEqual(mock_queue.call_count, 1)


class TestFormats(unittest.TestCase):
    """Tests for the supported formats."""

    def setUp(self):
        self.p = Project(init=False)

    def test_extensions(self):
        """Test returning the correct extension for a format."""
        sample_formats = {
            'PO': {'file-extensions': '.po, .pot'},
            'QT': {'file-extensions': '.ts'},
        }
        extensions = ['.po', '.ts', '', ]
        with patch.object(self.p, "do_url_request") as mock:
            mock.return_value = json.dumps(sample_formats), "utf-8"
            for (type_, ext) in zip(['PO', 'QT', 'NONE', ], extensions):
                extension = self.p._extension_for(type_)
                self.assertEqual(extension, ext)


class TestOptions(unittest.TestCase):
    """Test the methods related to parsing the configuration file."""

    def setUp(self):
        self.p = Project(init=False)

    def test_get_option(self):
        """Test _get_option method."""
        with patch.object(self.p, 'get_resource_option') as rmock:
            with patch.object(self.p, 'config', create=True) as cmock:
                rmock.return_value = 'resource'
                cmock.has_option.return_value = 'main'
                cmock.get.return_value = 'main'
                self.assertEqual(self.p._get_option(None, None), 'resource')
                rmock.return_value = None
                cmock.has_option.return_value = 'main'
                cmock.get.return_value = 'main'
                self.assertEqual(self.p._get_option(None, None), 'main')
                cmock.has_option.return_value = None
                self.assertIs(self.p._get_option(None, None), None)


class TestConfigurationOptions(unittest.TestCase):
    """Test the various configuration options."""

    def test_i18n_type(self):
        p = Project(init=False)
        i18n_type = 'PO'
        with patch.object(p, 'config', create=True) as config_mock:
            p.set_i18n_type([], i18n_type)
            calls = config_mock.method_calls
            self.assertEqual('set', calls[0][0])
            self.assertEqual('main', calls[0][1][0])
            p.set_i18n_type(['transifex.txo'], 'PO')
            calls = config_mock.method_calls
            self.assertEqual('set', calls[0][0])
            p.set_i18n_type(['transifex.txo', 'transifex.txn'], 'PO')
            calls = config_mock.method_calls
            self.assertEqual('set', calls[0][0])
            self.assertEqual('set', calls[1][0])


class TestStats(unittest.TestCase):
    """Test the access to the stats objects."""

    def setUp(self):
        self.stats = Mock()
        self.stats.__getitem__ = Mock()
        self.stats.__getitem__.return_value = '12%'

    def test_field_used_per_mode(self):
        """Test the fields used for each mode."""
        Project._extract_completed(self.stats, 'translate')
        self.stats.__getitem__.assert_called_with('completed')
        Project._extract_completed(self.stats, 'reviewed')
        self.stats.__getitem__.assert_called_with('reviewed_percentage')
